import abc

from blacksmith.domain.model import HTTPRequest, HTTPResponse, HTTPTimeout
from blacksmith.typing import HttpMethod, Proxies


class AsyncAbstractTransport(abc.ABC):
    def __init__(self, verify_certificate: bool = True, proxies: Proxies = None):
        self.verify_certificate = verify_certificate
        self.proxies = proxies

    @abc.abstractmethod
    async def request(
        self, method: HttpMethod, request: HTTPRequest, timeout: HTTPTimeout
    ) -> HTTPResponse:
        """Process the HTTP Get request."""



import httpx
from httpx import Response as HttpxRepsonse
from httpx import Timeout as HttpxTimeout

from blacksmith.domain.exceptions import HTTPError, HTTPTimeoutError
from blacksmith.domain.model import HTTPRequest, HTTPResponse, HTTPTimeout
from blacksmith.service.ports import SyncClient
from blacksmith.typing import HttpMethod

from ..base import SyncAbstractTransport


def safe_json(r: HttpxRepsonse):
    try:
        return r.json()
    except Exception:
        return {"error": r.text}


class SyncHttpxTransport(SyncAbstractTransport):
    """
    Transport implemented using `httpx`_.

    .. _`httpx`: https://www.python-httpx.org/

    """

    def request(
        self, method: HttpMethod, request: HTTPRequest, timeout: HTTPTimeout
    ) -> HTTPResponse:
        headers = request.headers.copy()
        if request.body:
            headers["Content-Type"] = "application/json"
        with SyncClient(verify=self.verify_certificate, proxies=self.proxies) as client:
            try:
                r = client.request(
                    method,
                    request.url,
                    params=request.querystring,
                    headers=headers,
                    content=request.body,
                    timeout=HttpxTimeout(timeout.request, connect=timeout.connect),
                )
            except httpx.TimeoutException as exc:
                raise HTTPTimeoutError(
                    f"{exc.__class__.__name__} while calling {method} {request.url}"
                )

        json = "" if r.status_code == 204 else safe_json(r)
        resp = HTTPResponse(r.status_code, r.headers, json=json)
        if not r.is_success:
            raise HTTPError(f"{r.status_code} {r.reason_phrase}", request, resp)
        return resp



"""Cut the circuit in case a service is down."""

from typing import Any, Iterable, Optional, cast

from purgatory import SyncAbstractUnitOfWork, SyncCircuitBreakerFactory
from purgatory.typing import TTL, Hook, Threshold

from blacksmith.domain.exceptions import HTTPError
from blacksmith.domain.model.http import HTTPRequest, HTTPResponse
from blacksmith.typing import ClientName, HttpMethod, Path

from .base import SyncHTTPMiddleware, SyncMiddleware
from .prometheus import SyncPrometheusMetrics

Listeners = Optional[Iterable[Hook]]


def exclude_httpx_4xx(exc: HTTPError):
    """Exclude client side http errors."""
    return exc.is_client_error


class GaugeStateValue:
    CLOSED = 0
    HALF_OPEN = 1
    OPEN = 2


class PrometheusHook:
    def __init__(self, prometheus_metrics: SyncPrometheusMetrics):
        self.prometheus_metrics = prometheus_metrics

    def __call__(self, circuit_name: str, evt_type: str, payload: Any) -> None:
        if evt_type == "state_changed":
            state = {
                "closed": GaugeStateValue.CLOSED,
                "half-opened": GaugeStateValue.HALF_OPEN,
                "opened": GaugeStateValue.OPEN,
            }[payload.state]
            self.prometheus_metrics.blacksmith_circuit_breaker_state.labels(
                circuit_name
            ).set(state)
        elif evt_type == "failed":
            self.prometheus_metrics.blacksmith_circuit_breaker_error.labels(
                circuit_name
            ).inc()


class SyncCircuitBreaker(SyncHTTPMiddleware):
    """
    Prevent cascading failure.

    The circuit breaker is based on `purgatory`_, the middleware create
    one circuit breaker per client_name. The parameters are forwarded
    to all the clients. This middleware does not give the possibility to
    adapt a threshold or the time the circuit is opened per clients.

    .. _`purgatory`: https://pypi.org/project/purgatory-circuitbreaker/
    """

    def __init__(
        self,
        threshold: Threshold = 5,
        ttl: TTL = 30,
        listeners: Listeners = None,
        uow: Optional[SyncAbstractUnitOfWork] = None,
        prometheus_metrics: Optional[SyncPrometheusMetrics] = None,
    ):
        self.circuit_breaker = SyncCircuitBreakerFactory(
            default_threshold=threshold,
            default_ttl=ttl,
            exclude=[(HTTPError, exclude_httpx_4xx)],
            uow=uow,
        )
        if prometheus_metrics:
            self.circuit_breaker.add_listener(PrometheusHook(prometheus_metrics))
        if listeners:
            for listener in listeners:
                self.circuit_breaker.add_listener(listener)

    def initialize(self):
        self.circuit_breaker.initialize()

    def __call__(self, next: SyncMiddleware) -> SyncMiddleware:
        def handle(
            req: HTTPRequest, method: HttpMethod, client_name: ClientName, path: Path
        ) -> HTTPResponse:

            with self.circuit_breaker.get_breaker(client_name):
                resp = next(req, method, client_name, path)
            return cast(HTTPResponse, resp)

        return handle



from typing import Any, Coroutine

from typing_extensions import Protocol

from blacksmith.domain.model.http import HTTPRequest, HTTPResponse
from blacksmith.typing import ClientName, HttpMethod, Path


class AsyncMiddleware(Protocol):
    """Signature of the middleware for the async version."""

    def __call__(
        self, req: HTTPRequest, method: HttpMethod, client_name: ClientName, path: Path
    ) -> Coroutine[Any, Any, HTTPResponse]:
        """This is the next function of the middleware."""


class SyncMiddleware(Protocol):
    """Signature of the middleware for the sync version."""

    def __call__(
        self, req: HTTPRequest, method: HttpMethod, client_name: ClientName, path: Path
    ) -> HTTPResponse:
        """This is the next function of the middleware."""



"""
Convenient method to fillout an blacksmith registry.

::

  import blacksmith
  blacksmith.scan('mypkg.resources', 'other.resources')
"""
import importlib
import pkgutil


def scan(*modules: str):
    """
    Collect all resources to fillout the registry.

    Basically, it import modules containins :func:`blacksmith.register` calls.

    """
    for modname in modules:
        mod = importlib.import_module(modname)
        for _loader, submod, _is_pkg in pkgutil.walk_packages(
            path=mod.__path__,
            prefix=mod.__name__ + ".",
        ):
            importlib.import_module(submod)
